"""repository abstraction for a mercurial repository

:organization: Logilab
:copyright: 2007-2013 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

from __future__ import with_statement

__docformat__ = "restructuredtext en"

import urllib2
from os import umask, makedirs
from os.path import split, exists, dirname
from datetime import datetime
from contextlib import contextmanager
import time

from mercurial import (hg, ui, context, util, encoding,
                       commands as hgcmd, phases, obsolete)
from mercurial.node import nullid, short as short_hex
from mercurial import scmutil

from logilab.common import tempattr

from cubicweb import Binary, QueryError, ValidationError, ExecutionError
from cubicweb.server.ssplanner import EditedEntity

from cubes.vcsfile import repo, bridge
from cubes.vcsfile.entities import remove_authinfo

_DELETED = object()

def monkeypatch_mercurial_url():
    from mercurial.url import httphandler
    if hasattr(httphandler, '__del__'):
        del httphandler.__del__
monkeypatch_mercurial_url()

def spop(aset, item):
    try:
        aset.remove(item)
    except KeyError:
        pass

@contextmanager
def force_mercurial_utf8():
    old_encoding = encoding.encoding
    encoding.encoding = 'utf-8'
    yield
    encoding.encoding = old_encoding

# mercurial 2.4 at least does not contain any code to get
# the precursors ...
def iter_direct_known_precursors(ctx):
    obsstore = getattr(ctx._repo, 'obsstore', None)
    if obsstore is None:
        return
    nodemap = ctx._repo.changelog.nodemap
    markers = obsstore.precursors.get(ctx.node(), ())
    candidates = set(mark[0] for mark in markers)
    seen = set(candidates)
    # we traverse chains of markers which may refer to
    # nodes not known from the repo (because they were
    # not pushed/pulled)
    while candidates:
        current = candidates.pop()
        crev = nodemap.get(current)
        # is the marked revision known from the repo ?
        if crev is not None:
            yield ctx._repo[crev]
            continue
        # if not, let's continue until we reach a known precursor
        for mark in obsstore.precursors.get(current, ()):
            if mark[0] not in seen:
                candidates.add(mark[0])
                seen.add(mark[0])

# mercurial 2.4 at least does not contain any code to get
# the precursors ...
def iter_direct_known_successors(ctx):
    obsstore = getattr(ctx._repo, 'obsstore', None)
    if obsstore is None:
        return
    nodemap = ctx._repo.changelog.nodemap
    markers = obsstore.successors.get(ctx.node(), ())
    candidates = set()
    for mark in markers:
        candidates.update(set(mark[1]))
    seen = set(candidates)
    # we traverse chains of markers which may refer to
    # nodes not known from the repo (because they were
    # not pushed/pulled)
    while candidates:
        current = candidates.pop()
        crev = nodemap.get(current)
        # is the marked revision known from the repo ?
        if crev is not None:
            yield ctx._repo[crev]
            continue
        # if not, let's continue until we reach a known precursor
        for mark in obsstore.precursors.get(current, ()):
            for succ in mark[1]:
                if succ not in seen:
                    candidates.add(succ)
                    seen.add(succ)

class vcsfile_ui(ui.ui):
    """ redirect output to cubicweb log """
    def __init__(self, loggable):
        ui.ui.__init__(self)
        self.cw_loggable = loggable

    def copy(self):
        return self

    def interactive(self):
        return False

    # XXX beware encoding issues wrt localized outputs ...
    def write(self, *args, **kwargs):
        if self._buffers:
            self._buffers[-1].extend([str(a) for a in args])
        else:
            for a in args:
                self.cw_loggable.debug(str(a).strip())

    def write_err(self, *args, **kwargs):
        for a in args:
            self.cw_loggable.error(str(a).strip())


def bw_nb_revisions(changelog):
    """return the number of revisions in a repository's changelog,
    assuring mercurial < 1.1 compatibility
    """
    try:
        return len(changelog)
    except AttributeError: # hg < 1.1
        return changelog.count()


class HGRepository(repo.VCSRepository):
    type = 'hg'

    def pull_or_clone_cache(self, url):
        vui = vcsfile_ui(self)
        if not exists(self.path):
            pdir = dirname(self.path)
            if not exists(pdir):
                makedirs(pdir)
            self.info('cloning %s into %s', url, self.path)
            try:
                hgcmd.clone(vui, str(url),
                            dest=self.path, stream='compressed', noupdate=True)
            except (util.Abort, urllib2.URLError), ex:
                raise bridge.VCSException(self.eid, 'url', '%s (%s)', (ex, url))
        else:
            self.debug('pulling from %s into %s', remove_authinfo(url), self.path)
            try:
                hgcmd.pull(vui, hg.repository(vui, self.path), source=str(url))
            except (util.Abort, urllib2.URLError), ex:
                raise bridge.VCSException(self.eid, 'path', '%s (%s)', (ex, url,))
        if obsolete._enabled:
            # Wait 15s after a pull and re-pull the null revision, to make it
            # less likely we're missing in-flight obsolescence markers.
            # see #2818315
            time.sleep(15)
            try:
                hgcmd.pull(vui, hg.repository(vui, self.path), source=str(url), rev=('null',))
            except (util.Abort, urllib2.URLError), ex:
                raise bridge.VCSException(self.eid, 'path', '%s (%s)', (ex, url))

    def import_content(self, repoentity, commitevery=1000):
        """import content from the mercurial repository"""
        assert self.eid == repoentity.eid
        if not exists(self.path):
            self.pull_or_clone_cache(repoentity.source_url)

        with force_mercurial_utf8():
            try:
                repo = self.hgrepo()
                # mercurial 2.5
                if hasattr(repo, 'unfiltered'):
                    repo = repo.unfiltered()
            except Exception, ex:
                self.exception('Impossible to open repository %s (%s)', self, ex)
                return

            session = repoentity._cw
            knownheads = [(str(r.changeset), r.revision)
                          for r in repoentity.heads_rset().entities()]

            # detect stripped branches
            needsstrip = False
            needsrenumbering = False
            for cset, rev in knownheads:
                if cset not in repo:
                    needsstrip = True
                    self.warning('%s: strip detected. %s is unknown in local cache',
                                 self.path, cset)
            if not needsstrip:
                incoming = repo.revs("not ::%ls", [cset for cset, _rev in knownheads])
                if incoming:
                    needsrenumbering = session.execute('Any R WHERE R is Revision, '
                                                       'R from_repository REPO, '
                                                       'REPO eid %%(reid)s, R revision IN (%s)'
                                                       % ','.join([str(rev) for rev in incoming]),
                                                       {'reid': repoentity.eid})
                if not needsrenumbering:
                    for cset, rev in knownheads:
                        if repo.revs("id(%s)" % str(cset)) != [rev]:
                            needsrenumbering = True
                            break

            if needsstrip or needsrenumbering:
                self.warning('%s: strip and renumbering from %s.',
                             repoentity.dc_title(), self.path)
                # clean all 'revision' attribute to please unicity
                # constraint when renumbering
                session.execute('SET R revision 0-REV WHERE R revision REV, R from_repository REPO, '
                                'REPO eid %(reid)s', {'reid': repoentity.eid})
                revs = session.execute('Any R WHERE R is Revision, R from_repository REPO, '
                                       'REPO eid %(reid)s', {'reid': repoentity.eid})
                for reventity in revs.entities():
                    if str(reventity.changeset) not in repo:
                        reventity.cw_delete()
                    else:
                        rev = repo[str(reventity.changeset)].rev()
                        reventity.cw_set(revision=rev)

                if session.execute('Revision R WHERE R revision < 0, R from_repository REPO, '
                                       'REPO eid %(reid)s', {'reid': repoentity.eid}):
                    session.rollback()
                    self.error('Abort stripping and renumbering of Repository %s: '
                               'detected remaining Revision with negative revision number',
                               repoentity.eid)
                    return
                session.commit()
                session.set_cnxset()
                session.execute(
                    'DELETE VersionedFile X WHERE X from_repository R, R eid %(r)s, '
                    'NOT VC content_for X', {'r': repoentity.eid})
                # commit and give a chance to threads waiting for a cnxset
                # by freeing it then reacquiring it
                session.commit()
                session.set_cnxset()

            # update phases and visibility of already known revisions
            self.update_phases(session, repoentity, repo)
            self.update_hidden(session, repoentity, repo)
            # XXX we need to update obsolescence markers without new revision
            #     involved.
            knownheads = [str(r.changeset) for r in repoentity.heads_rset().entities()]
            missing = repo.revs("not ::%ls", knownheads)
            missing.sort()
            while missing:
                for rev in missing[:commitevery]:
                    try:
                        self.import_revision(repoentity, repo, rev)
                    except Exception:
                        self.critical('error while importing revision %s of %s',
                                      rev, self.path, exc_info=True)
                        raise
                # give a chance to threads waiting for a cnxset by
                # freeing it then reacquiring it
                session.commit()
                session.set_cnxset()
                del missing[:commitevery] # pop iterated value
            # if missing was empty we still need to commit phases and visibility changes
            # give a chance to threads waiting for a cnxset by
            # freeing it then reacquiring it
            session.commit()
            session.set_cnxset()

    def update_phases(self, session, repoentity, repo):
        """ update all phases (this potentially affects all revs since
        phase changes are not transactional) """
        notpublic = set(str(ctx) for ctx in repo.set('not public()'))
        if notpublic:
            session.execute('SET R phase "public" WHERE NOT R changeset IN (%s),'
                            'NOT R phase "public", R from_repository REPO,'
                            'REPO eid %%(repo)s' %
                            ','.join(repr(rev) for rev in notpublic),
                            {'repo': repoentity.eid})
            for phase in ('draft', 'secret'):
                revs = set(str(ctx) for ctx in repo.set('%s()' % phase))
                if revs:
                    session.execute('SET R phase "%s" WHERE R changeset IN (%s),'
                                    'NOT R phase "%s", R from_repository REPO,'
                                    'REPO eid %%(repo)s' %
                                    (phase, ','.join(repr(rev) for rev in revs), phase),
                                    {'repo': repoentity.eid})
        else:
            session.execute('SET R phase "public" WHERE R from_repository REPO, '
                            'NOT R phase "public", REPO eid %(repo)s',
                            {'repo': repoentity.eid})



    def update_hidden(self, session, repoentity, repo):
        """ maintain hidden status for all revs """
        oldhidden = set(num for num, in session.execute('Any N WHERE R hidden True, '
                                                        'R from_repository H, '
                                                        'R revision N, H eid %(hg)s',
                                                        {'hg': repoentity.eid}).rows)
        newhidden = set(repo.revs('hidden()'))
        hide = newhidden - oldhidden
        show = oldhidden - newhidden
        for revs, setto in ((hide, 'True'), (show, 'False')):
            if revs:
                session.execute('SET R hidden %s WHERE R from_repository H,'
                                'H eid %%(hg)s, R revision IN (%s)' %
                                (setto, ','.join(str(revnum) for revnum in revs)),
                                {'hg': repoentity.eid})

        # very partial handling of obsolescence relation added after a revision
        # issue #2731056
        # this tries to find explanation for changeset that becomes hidden
        # - does not work for changesets already hidden
        # - does not work for public changesets
        for new_hide in hide:
            ctx = repo[new_hide]
            RQL = '''SET S obsoletes P
                     WHERE R eid %(r)s,
                           P from_repository R,
                           P changeset %(p)s,
                           S from_repository R,
                           S changeset %(s)s,
                           NOT S obsoletes P
            '''
            # unknown successors will create the relation at insertion time
            data = {'r': repoentity.eid,
                    'p': str(ctx)}
            for succ in iter_direct_known_successors(ctx):
                data['s'] = str(succ)
                session.execute(RQL, data)


    def import_revision(self, repoentity, repo, i):
        self.info('importing revision %s from %s', i, repoentity.dc_title())
        session = repoentity._cw
        execute = session.execute
        node = repo.changelog.node(i)
        changeset = unicode(short_hex(node))
        if execute('Any X WHERE X from_repository R, R eid %(repo)s, X changeset %(cs)s',
                   {'repo': repoentity.eid, 'cs': changeset}):
            self.warning('skip revision %s, seems already imported',
                         changeset)
            return
        ctx = repo.changectx(i)
        date = datetime.fromtimestamp(ctx.date()[0])
        #taglist = ctx.tags() #repo.nodetags(node)
        revdata = {'date': date, 'revision': i,
                   'author': unicode(ctx.user(), 'utf-8'),
                   'description': unicode(ctx.description(), 'utf-8'),
                   'changeset': changeset,
                   'phase': unicode(ctx.phasestr()),
                   'branch': unicode(ctx.branch()),
                   'hidden': ctx.hidden()
                   }
        parent_changesets = [short_hex(n)
                             for n in repo.changelog.parents(node)
                             if n != nullid]
        precursors_changesets = [str(p) for p in iter_direct_known_precursors(ctx)]
        if not precursors_changesets:
            precursors = ()
        else:
            precursors = execute(
                'Any X WHERE X revision XR, X changeset XC, '
                'X changeset IN (%s), X from_repository R, R eid %%(r)s'
                % ','.join("'%s'" % cs for cs in precursors_changesets),
                {'r': self.eid})
        revdata['precursors'] = [r[0] for r in precursors]
        if not parent_changesets:
            parents = ()
        elif len(parent_changesets) == 1:
            parents = execute(
                'Any X,XC,XR WHERE X revision XR, X changeset XC, '
                'X changeset %(cs)s, X from_repository R, R eid %(r)s',
                {'cs': parent_changesets[0], 'r': self.eid})
        else:
            parents = execute(
                'Any X,XC,XR WHERE X revision XR, X changeset XC, '
                'X changeset IN (%s), X from_repository R, R eid %%(r)s'
                % ','.join("'%s'"%cs for cs in parent_changesets),
                {'r': self.eid})
        revdata['parents'] = [r[0] for r in parents]
        reveid = bridge.import_revision(session, self.eid, **revdata)
        if not repoentity.import_revision_content:
            return
        changes = repo.status(ctx.parents()[0].node(), ctx.node())[:3]
        modified, added, removed = changes
        path_filter = repoentity.subpath
        for path in modified + added:
            upath = unicode(path, self.encoding, 'replace')
            if not (path_filter is None or upath.startswith(path_filter)):
                continue
            self._import_version_content(session, changeset, reveid, removed,
                                         path, upath, date, repo, ctx)
        for path in removed:
            upath = unicode(path, self.encoding, 'replace')
            if not (path_filter is None or upath.startswith(path_filter)):
                continue
            bridge.import_deleted_version_content(session, self.eid, reveid,
                                                  upath, date)
        parent_idx = dict( ((cs, lrev) for _, cs, lrev in parents) )
        parent_lrevs = [parent_idx[cs] for cs in parent_changesets]
        bridge.set_at_revision(session, reveid, parent_lrevs)

    def _import_version_content(self, session, changeset, reveid, removed, path, upath,
                                date, repo, ctx):
        directory, fname = split(path)
        session.transaction_data['vcs_importing'] = self.eid, directory, fname, changeset
        filectx = ctx[path]
        vcdata = {'data': Binary(filectx.data())}
        renameinfo = filectx.renamed()
        if renameinfo:
            pvc, oldfile = self._renamed_vc_rset(session, repo, renameinfo)
            if pvc:
                assert len(pvc) == 1
                if oldfile in removed:
                    vcdata['vc_rename'] = pvc[0][0]
                else:
                    vcdata['vc_copy'] = pvc[0][0]
            else:
                self.error('detected copy or rename of %s@%s but unable'
                           ' to find associated version content',
                           oldfile, pvc.args['cs'])
        return bridge.import_version_content(session, self.eid, reveid, upath,
                                             date, **vcdata)

    def _renamed_vc_rset(self, session, repo, renameinfo):
        oldfile, fileid = renameinfo
        pfctx = repo.filectx(oldfile, fileid=fileid)
        pcs = short_hex(pfctx.node())
        dir, name = split(unicode(oldfile, self.encoding, 'replace'))
        return session.execute('VersionContent X WHERE '
                               'X from_revision REV, REV changeset %(cs)s, '
                               'REV from_repository R, R eid %(r)s, '
                               'X content_for VF, VF directory %(dir)s, VF name %(name)s',
                               {'cs':  pcs, 'r': self.eid,
                                'dir': dir, 'name': name}), oldfile

    def hgrepo(self):
        repo = hg.repository(ui.ui(), self.path)
        return getattr(repo, 'unfiltered', lambda: repo)()

    def _file_content(self, path, rev):
        """extract a binary string with the content of the file at `path` for
        revision `rev` in the repository
        """
        ctx = self.hgrepo().changectx(rev)
        return Binary(ctx[path].data())

    def revision_transaction(self, session, entity):
        """open a transaction to create a new revision corresponding to the
        given entity
        """
        return HGTransaction(self, session, entity)

    def add_versioned_file_content(self, session, transaction, vf, entity,
                                   data):
        """add a new revision of a versioned file"""
        vfpath = self.encode(vf.path)
        transaction.changes[vfpath] = context.memfilectx(
            vfpath, data.getvalue(), islink=False, isexec=False, copied=None)
        return vf.path

    def add_versioned_file_deleted_content(self, session, transaction, vf,
                                           entity):
        """add a new revision of a just deleted versioned file"""
        transaction.changes[self.encode(vf.path)] = _DELETED
        return vf.path


class HGTransaction(object):
    def __init__(self, repohdlr, session, revision):
        self.repohdlr = repohdlr
        self.session = session
        self.revision = revision
        self.repo = repohdlr.hgrepo()
        # see http://mercurial.selenic.com/wiki/LockingDesign
        self._lock = self.repo.lock()
        self.extra = {}
        self.changes = {}

    @property
    def rev(self):
        """newly created revision number"""
        return bw_nb_revisions(self.repo.changelog)

    def _filectx(self, repo, memctx, path):
        """callable receiving the repository, the current memctx object and the
        normalized path of requested file, relative to repository root. It is
        fired by the commit function for every file in 'files', but calls order
        is undefined. If the file is available in the revision being committed
        (updated or added), filectxfn returns a memfilectx object. If the file
        was removed, filectxfn raises an IOError. Moved files are represented by
        marking the source file removed and the new file added with copy
        information (see memfilectx).
        """
        if self.changes[path] is _DELETED:
            raise IOError()
        return self.changes[path]

    def precommit(self):
        if not self.changes:
            raise QueryError('nothing changed')
        # XXX to what commit does here + strip in revert_precommit
        branch = self.repohdlr.encode(
            self.revision.cw_attr_cache.get('branch') or u'default')
        if branch != u'default':
            self.extra['branch'] = branch
        try:
            self.hgparent = p1 = self.repo.branchtags()[branch]
        except KeyError:
            # new branch
            self.hgparent = p1 = self.repo.branchtags().get('default')
        if p1:
            if not self.revision.parent_revision:
                raise ValidationError(self.revision.eid,
                                      {'parent_revision': 'missing expected parent'})
            if [short_hex(p1)] != [r.changeset for r in self.revision.parent_revision]:
                raise ValidationError(None, {None: ('concurrency error, please '
                                                    're-upload the revision')})

    def commit(self):
        # XXX: make umask configurable
        oldumask = umask(022)
        try:
            self._commit()
        finally:
            umask(oldumask)

    def _commit(self):
        # XXX merging branches
        repohdlr = self.repohdlr
        revision = self.revision
        author = repohdlr.encode(revision.cw_attr_cache.get('author', u''))
        msg = repohdlr.encode(revision.cw_attr_cache.get('description', u''))
        # ensure mercurial will use configured repo encoding, not locale's
        # encoding
        # XXX modifying module's global is not very nice but I've no other idea
        old_encoding = encoding.encoding
        try:
            encoding.encoding = repohdlr.encoding
            ctx = context.memctx(self.repo, (self.hgparent, None), msg,
                                 self.changes.keys(), self._filectx, author,
                                 extra=self.extra)
            node = self.repo.commitctx(ctx)
            # update revision's record to set correct changeset
            #
            # take care, self.repo != self.session.repo (eg mercurial repo
            # instance / rql repository)
            source = self.session.repo.system_source
            # remove previous cached value to only save the changeset
            with tempattr(revision, 'cw_edited', EditedEntity(revision)) as rev:
                rev.cw_edited['changeset'] = unicode(short_hex(node))
                rev.cw_edited['phase'] = unicode(self.repo[node].phasestr())
                source.update_entity(self.session, rev)
            # we need to explicitly commit the connection to the database since
            # this method is actually called from a postcommit event, i.e. once
            # cnxset has already been commited
            self.session.cnxset.connection('system').commit()
            # XXX restore entity's dict?
            self.rollback()
        finally:
            encoding.encoding = old_encoding

    def rollback(self):
        try:
            from mercurial.lock import release
            release(self._lock)
        except ImportError:
            try:
                # mercurial >= 1.3.1
                self._lock.release()
            except AttributeError:
                del self._lock


from logging import getLogger
from cubicweb import set_log_methods
set_log_methods(HGRepository, getLogger('cubicweb.sources.hg'))



"""app objects for vcsfile web interface

:organization: Logilab
:copyright: 2007-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""

__docformat__ = "restructuredtext en"


from cubicweb import typed_eid
from cubicweb.selectors import match_form_params
from cubicweb.web import Redirect, RequestError
from cubicweb.web.controller import Controller


class RefreshLocalRepoCachesController(Controller):
    """Control used to refresh the local cache of a single vcs repository given
    its eid in the web request form"""
    __regid__ = 'refresh_repository'
    __select__ = Controller.__select__ & match_form_params('eid')

    def publish(self, rset=None):
        eid = typed_eid(self._cw.form['eid'])
        vcsrepo = self._cw.entity_from_eid(eid)
        if vcsrepo.e_schema.type != 'Repository':
            raise RequestError('a repository is expected')
        # async=False so that users see changes immediately, if any
        self._cw.cnx.call_service('refresh_repository', async=False,
                                  eids=(vcsrepo.eid,))
        raise Redirect(vcsrepo.absolute_url())



# repository specific stuff ####################################################

try:
    from cubicweb import server
except ImportError: # no server installation
    pass
else:
    options = (
        ('repository-import',
         {'type' : 'yn',
          'default': True,
          'help': 'Is the instance responsible to automatically import new '
                   'revisions from repositories? '
                   'You should say yes unless you don\'t want this behaviour '
                   'or if you use a multiple repositories setup, in which '
                   'case you should say yes on one repository, no on others.',
          'level': 2,
          'group': 'vcsfile',
          }),
        ('check-revision-interval',
         {'type' : 'int',
          'default': 5*60,
          'help': 'interval between checking of new revisions in repositories \
(default to 5 minutes).',
          'level': 2,
          'group': 'vcsfile',
          }),
        ('check-revision-commit-every',
         {'type' : 'int',
          'default': 1000,
          'help': 'after how much new imported revisions the transaction \
should be commited (default to 1000).',
          'level': 2,
          'group': 'vcsfile',
          }),
        ('import-revision-content',
         {'type' : 'yn',
          'default': True,
          'help': 'This option is now deprecated in favor of the import_revision_content attribute per Repository.',
          'level': 2,
          'group': 'vcsfile',
          }),
        ('local-repo-cache-root',
         {'type':'string',
          'default': 'repo_cache', # XXX /var/lib/cubicweb/<instance>
          'help':'Local repository cache location (if not absolute, will be '
          'relative to instance data home directory).',
          'level': 2,
          'group': 'vcsfile'
          }
         ),
        ('cache-external-repositories',
         {'type':'yn',
          'default': False,
          'help':'Should repositories from external source have a local cache?'
          'This is usually not necessary, beside cases where for instance a '
          'narval bot on the same host as the instance could benefit from them '
          'to run apycot tests.',
          'level': 2,
          'group': 'vcsfile'
          }
         ),
        ('hgrc-path',
         {'type': 'string',
          'default': None,
          'help': 'a list of files or directories to search for mercurial configuration',
          'level': 2,
          'group': 'vcsfile',
          }
         ),
        )

from cubicweb.mttransforms import ENGINE
from cubes.vcsfile.docparser import Diff2HTMLTransform

COMPONENT_CONTEXT = 'navcontentbottom'

if Diff2HTMLTransform.name not in ENGINE.transforms:
    ENGINE.add_transform(Diff2HTMLTransform())



from mercurial import cmdutil
from mercurial import patch
from mercurial import ui as uimod

from cubicweb import Binary
from cubicweb.server import Service

from . import bridge


class RevisionExportService(Service):
    """return the patch version of a revision
    """

    __regid__ = 'vcs.export-rev'

    def call(self, repo, nodeid):
        repo = self._cw.entity_from_eid(repo)
        if repo.type != 'mercurial':
            # no svn support yet (use a selector when we have multiple versions)
            return None
        ui = uimod.ui()
        diffopts = patch.diffopts(ui, {'git': True,
                                       'showfunc': True,
                                       'unified': 5,
                                       })
        hdrepo = bridge.repository_handler(repo)
        hgrepo = hdrepo.hgrepo()
        if nodeid not in hgrepo:
            # local cache miss
            return None
        revs = [nodeid]
        output = Binary()
        cmdutil.export(hgrepo, revs, fp=output, opts=diffopts)
        return output.getvalue()



"""entity types used to represent a version control system (vcs) content

:organization: Logilab
:copyright: 2007-2011 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
__docformat__ = "restructuredtext en"
_ = unicode

from yams.buildobjs import (DEFAULT_ATTRPERMS,
                            EntityType, RelationDefinition,
                            String, Int, Boolean, Bytes)
from cubicweb.schema import RRQLExpression

PRIVATE_ATTRPERMS = DEFAULT_ATTRPERMS.copy()
PRIVATE_ATTRPERMS['read'] = ('managers',)
PRIVATE_ATTRPERMS['update'] = ('managers',)

RO_ATTRPERMS = DEFAULT_ATTRPERMS.copy()
RO_ATTRPERMS['update'] = ()
RO_RELPERMS = DEFAULT_ATTRPERMS.copy()
RO_RELPERMS['add'] = (RRQLExpression('X from_repository R, U in_group G, '
                                     'R require_permission P, P require_group G'))


class Repository(EntityType):
    """a local subversion repository which will be used as versionned content
    entity source
    """
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers',),
        'update': ('managers',),
        'delete': ('managers',),
        }

    title = String(description=_('optional title. If the repository has a public url, '
                                 'it may be a good idea to put it as title.'),
                   maxsize=256)
    type = String(required=True, vocabulary=('subversion', 'mercurial'),
                  description=_('repository\'s type'))
    source_url = String(description=_('source url (must be of the form protocol://path/to/stuff)'),
                        unique=True)
    path = String(unique=True,
                  description=_('absolute path to the local repository'),
                  __permissions__=PRIVATE_ATTRPERMS)
    local_cache = String(unique=True,
                         description=_('relative path to the local cache of the repository'),
                         __permissions__=PRIVATE_ATTRPERMS)
    subpath = String(description=_('if specified, only import the given subpart'
                                   ' of the repository (path relative to the '
                                   'repository root)'))
    encoding = String(default='utf-8', maxsize=20, required=True,
                      description=_('encoding used for the repository (e.g.'
                                    ' for file path and check-in comments)'))
    import_revision_content = Boolean(default=True,
                                      description=_('import contents of files touched by a revision. '
                                                    'It may cause performance issue when scaling.'))
    has_anonymous_access = Boolean(default=True,
                                   description=_('Tell if user needs an account to checkout the repository.'))


class Revision(EntityType):
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers', 'users',),
        'update': ('managers', 'owners'),
        'delete': (),
        }
    __unique_together__ = [('from_repository', 'revision'),
                           # XXX behaviour vs NULL values depending on the
                           # backend? NULL values are not considered equals by
                           # sqlite nor postgres at least
                           ('from_repository', 'changeset')]
    # NOTE: creation_date will be set at revision date
    revision = Int(required=True, indexed=True,
                   description=_('revision number.'),
                   __permissions__=RO_ATTRPERMS)

    description = String(description=_('comment for this revision.'))

    author = String(indexed=True, maxsize=200,
                    description=_("author of this revision."))

    changeset = String(indexed=True, maxsize=100,
                       description=_('change set identifier, when used by the '
                                     'underlying version control system.'),
                   __permissions__=RO_ATTRPERMS)

    branch = String(indexed=True, maxsize=255,
                    description=_("branch of this revision."),
                   __permissions__=RO_ATTRPERMS)

    phase = String(description=_('phase of the changeset'), indexed=True,
                   default='draft', vocabulary=[_('secret'), _('draft'), _('public')])

    hidden = Boolean(required=True, indexed=True)


class VersionedFile(EntityType):
    """a file stored in a versioned source"""
    __unique_together__ = [('from_repository', 'directory', 'name')]
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers', 'users',),
        'update': (),
        'delete': (),
        }
    directory = String(required=True, indexed=True,
                       description=_('directory of the file in the repository.'),
                       __permissions__=RO_ATTRPERMS)
    name = String(required=True, fulltextindexed=True,
                  description=_('name of the file in the repository.'),
                   __permissions__=RO_ATTRPERMS)


class DeletedVersionContent(EntityType):
    """represent a revision where a versioned file has been deleted"""
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    ('managers', 'users',),
        'update': ('managers', 'owners'),
        'delete': (),
        }


class VersionContent(DeletedVersionContent):
    """actual content of a versioned file, for a given revision"""
    data = Bytes(description=_("file's data"), required=True,
                 __permissions__=RO_ATTRPERMS)
    data_format = String(required=True, meta=True, maxsize=128,
                         description=_('MIME type of the file.'))
    data_encoding = String(meta=True, maxsize=32,
                           description=_('encoding of the file when it applies (e.g. text).'))


class from_repository(RelationDefinition):
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    (RRQLExpression('U in_group G, O require_permission P, '
                                  'P require_group G, P name "write"'),),
        'delete': (),
        }
    inlined = True
    subject = ('Revision', 'VersionedFile')
    object = 'Repository'
    cardinality = '1*'
    composite = 'object'

class content_for(RelationDefinition):
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    (RRQLExpression('O from_repository R, R require_permission P, '
                                  'U in_group G, P require_group G, P name "write"'),),
        'delete': (),
        }
    inlined = True
    subject = ('DeletedVersionContent', 'VersionContent')
    object = 'VersionedFile'
    composite = 'object'
    cardinality = '1*'

class from_revision(RelationDefinition):
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    (RRQLExpression('O from_repository R, R require_permission P, '
                                  'U in_group G, P require_group G, P name "write"'),),
        'delete': (),
        }
    inlined = True
    subject = ('DeletedVersionContent', 'VersionContent')
    object = 'Revision'
    cardinality = '1*'
    composite = 'object'

class parent_revision(RelationDefinition): # XXX ordered
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    (RRQLExpression('O from_repository R, R require_permission P, '
                                  'U in_group G, P require_group G, P name "write"'),),
        'delete': (),
        }
    subject = 'Revision'
    object = 'Revision'

class at_revision(RelationDefinition):
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    (),
        'delete': (),
        }
    subject = ('DeletedVersionContent', 'VersionContent')
    object = 'Revision'

class vc_copy(RelationDefinition):
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    (RRQLExpression('O content_for VF, VF from_repository R, R require_permission P, '
                                  'U in_group G, P require_group G, P name "write"'),),
        'delete': (),
        }
    inlined = True
    subject = object = 'VersionContent'
    cardinality = '?*'

class vc_rename(RelationDefinition):
    __permissions__ = {
        'read':   ('managers', 'users', 'guests'),
        'add':    (RRQLExpression('O content_for VF, VF from_repository R, R require_permission P, '
                                  'U in_group G, P require_group G, P name "write"'),),
        'delete': (),
        }
    inlined = True
    subject = object = 'VersionContent'
    cardinality = '?*'

class obsoletes(RelationDefinition):
    subject = 'Revision'
    object = 'Revision'

class granted_permission(RelationDefinition):
    subject = 'Repository'
    object = 'CWPermission'

class require_permission(RelationDefinition):
    subject = 'Repository'
    object = 'CWPermission'



import os.path as osp

from logilab.common.decorators import classproperty

from cubicweb.devtools.testlib import CubicWebTC

from cubes.vcsfile import bridge

def init_vcsrepo(repo, commitevery=1):
    bridge._REPOHDLRS.clear()
    session = repo.internal_session()
    try:
        bridge.import_content(session, commitevery=commitevery,
                              raise_on_error=True)
    finally:
        session.close()

class VCSRepositoryTC(CubicWebTC):
    """base class for test which should import a repository during setup"""
    _repo_path = None
    repo_type = u'mercurial'
    repo_encoding = u'utf-8'
    repo_subpath = None
    repo_import_revision_content = True
    repo_title = None

    commitevery = 3

    @classproperty
    def test_db_id(cls):
        if cls._repo_path is None:
            return None
        ids = [cls._repo_path]
        if cls.repo_subpath:
            ids.append(cls.repo_subpath)
        if not cls.repo_import_revision_content:
            ids.append('nocontent')
        return '-'.join(ids)

    @classproperty
    def repo_path(cls):
        assert cls._repo_path, 'you must set repository through _repo_path first'
        return osp.join(cls.datadir, cls._repo_path)

    @classmethod
    def _create_repo(cls, session):
        return session.create_entity(
            'Repository', type=cls.repo_type, path=cls.repo_path,
            subpath=cls.repo_subpath, encoding=cls.repo_encoding,
            title=cls.repo_title,
            import_revision_content=cls.repo_import_revision_content)

    @classmethod
    def grant_write_perm(cls, session, group, vcsrepoeid=None):
        if vcsrepoeid is None:
            vcsrepoeid = cls.vcsrepo.eid
        cls.grant_permission(session, vcsrepoeid, 'managers', 'write',
                             u'repo x write perm')

    @classmethod
    def pre_setup_database(cls, session, config):
        bridge._REPOHDLRS.clear()
        # don't use cls.vcsrepo in regular test, only in pre_setup_database
        cls.vcsrepo = cls._create_repo(session)
        session.commit()
        init_vcsrepo(session.repo, cls.commitevery)

    def setUp(self):
        bridge._REPOHDLRS.clear()
        super(VCSRepositoryTC, self).setUp()
        self.vcsrepo = self.execute('Repository X').get_entity(0, 0)



# -*- coding: utf-8 -*-
"""hooks for vcsfile content types

:organization: Logilab
:copyright: 2007-2012 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
:contact: http://www.logilab.fr/ -- mailto:contact@logilab.fr
"""
from __future__ import with_statement

__docformat__ = "restructuredtext en"

import os
import os.path as osp
import errno
import shutil

from logilab.mtconverter import need_guess, guess_mimetype_and_encoding

from cubicweb import QueryError, ValidationError
from cubicweb.server import hook, session, Service
from cubicweb.predicates import is_instance, score_entity, match_user_groups

from cubes.vcsfile import IMMUTABLE_ATTRIBUTES, bridge

def repo_cache_dir(config):
    """Return the directory to use for repo cache

    This function create directory if necessary."""
    directory = config['local-repo-cache-root']
    if not osp.isabs(directory):
        directory = osp.join(config.appdatahome, directory)
    config.local_repo_cache_root = directory
    if not osp.exists(directory):
        try:
            os.makedirs(directory)
        except Exception:
            config.critical('could not find local repo cache directory; check '
                            'local-repo-cache-root option whose present value '
                            'is %r)', config['local-repo-cache-root'])
            raise
    return directory

def url_to_relative_path(url):
    scheme, url = url.split('://', 1)
    if scheme == 'file':
        return url.rstrip('/').rsplit('/', 1)[-1]
    else:
        return url.rstrip('/').rsplit('/', 1)[1]

def set_local_cache(vcsrepo):
    cachedir = osp.join(repo_cache_dir(vcsrepo._cw.vreg.config),
                        unicode(vcsrepo.eid))
    if not osp.exists(cachedir):
        try:
            os.makedirs(cachedir)
        except OSError:
            vcsrepo.exception('cant create repo cache directory %s', cachedir)
            return
    try:
        vcsrepo.cw_edited['local_cache'] = osp.join(
            unicode(vcsrepo.eid), url_to_relative_path(vcsrepo.source_url))
    except (IndexError, ValueError):
        # do nothing, validation error should be raised by another hook
        pass

def clone_to_local_cache(vcsrepo):
    handler = bridge.repository_handler(vcsrepo)
    url = vcsrepo.source_url
    try:
        handler.pull_or_clone_cache(url)
    except Exception:
        handler.exception('while trying to clone repo %s', url)
        msg = vcsrepo._cw._('can not clone the repo from %s, '
                            'please check the source url')
        raise ValidationError(vcsrepo.eid, {'source_url': msg % url})


def missing_relation_error(entity, rtype):
    # use __ since msgid recorded in cw, we don't want to translate it in
    # this cube
    __ = entity._cw._
    msg = __('at least one relation %(rtype)s is required on %(etype)s (%(eid)s)')
    errors = {'from_repository': msg % {'rtype': __(rtype),
                                        'etype': __(entity.__regid__),
                                        'eid': entity.eid}}
    return ValidationError(entity.eid, errors)

def missing_attribute_error(entity, attrs):
    msg = _('at least one attribute of %s must be set on a Repository')
    errors = {}
    for attr in attrs:
        errors[attr] = msg % ', '.join(attrs)
    return ValidationError(entity.eid, errors)


# initialization hooks #########################################################


class VCSFileInitHook(hook.Hook):
    """install attribute map on the system source sql generator and init
    configuration
    """
    __regid__ = 'vcsfile.hook.init'
    events = ('server_startup', 'server_maintenance')

    def __call__(self):
        self.repo.system_source.set_storage('VersionContent', 'data',
                                       bridge.VCSStorage())
        cacheroot = repo_cache_dir(self.repo.config)
        if not os.access(cacheroot, os.W_OK):
            raise ValueError('directory %r is not writable (check "local-repo-'
                             'cache-root" option)' % cacheroot)



class VCSFileStartHook(hook.Hook):
    """register task to regularly pull/import vcs repositories content"""
    __regid__ = 'vcsfile.hook.start'
    events = ('server_startup',)

    def __call__(self):
        # closure to avoid error on code reloading
        interval = self.repo.config.get('check-revision-interval')
        if interval < 0:
            return
        def refresh_vcsrepos():
            return _refresh_repos(self.repo)
        self.repo.looping_task(interval, refresh_vcsrepos)

# internals to be able to create new vcs repo revision using rql queries #######

class VCSetAtRevisionOp(hook.Operation):

    def precommit_event(self):
        bridge.set_at_revision(self.session, self.revision.eid)


class VCTransactionOp(hook.LateOperation):

    def precommit_event(self):
        self.transaction.precommit()

    def revertprecommit_event(self):
        self.transaction.rollback()

    def postcommit_event(self):
        self.transaction.commit()

    def rollback_event(self):
        self.transaction.rollback()


class VFModificationDateOp(hook.DataOperationMixIn, hook.Operation):
    """set modification date of VersionedFile to the creation date of the
    latest revision modifying it
    """
    # use a list instead of a set as container to ensure revision order when
    # multiple revision are imported in a single transaction
    containercls = list

    def precommit_event(self):
        for entity in self.get_data():
            entity.file.cw_set(modification_date=entity.rev.creation_date)


class UpdateVFModificationDateHook(hook.Hook):
    __regid__ = 'vcsfile.update_vf_mdate'
    __select__ = hook.Hook.__select__ & is_instance('DeletedVersionContent',
                                                    'VersionContent')
    events = ('after_add_entity',)
    category = 'metadata'

    def __call__(self):
        VFModificationDateOp.get_instance(self._cw).add_data(self.entity)


class AddRevisionHook(hook.Hook):
    __regid__ = 'vcsfile.add_revision_hook'
    __select__ = hook.Hook.__select__ & is_instance('Revision')
    events = ('before_add_entity',)
    # no category, mandatory hook

    def __call__(self):
        entity = self.entity
        session = self._cw
        # new revision to be created, set a temporary value if necessary
        revision = entity.cw_edited.setdefault('revision', 0)
        # skip further processing if the revision is being imported from the
        # vcs repository
        if session.is_internal_session:
            return
        try:
            vcsrepoeid = entity.cw_attr_cache['from_repository']
        except KeyError:
            vcsrepoeid = session.transaction_data.pop('vcsrepoeid', None)
            if vcsrepoeid is None:
                raise missing_relation_error(entity, 'from_repository')
        if revision > 0: # set to 0 by hook
            raise QueryError("can't specify revision")
        transactions = session.transaction_data.setdefault('vctransactions', {})
        # should not have multiple transaction on the same repository
        if vcsrepoeid in transactions:
            raise QueryError('already processing a new revision')
        vcsrepohdlr = bridge.repository_handler(
            session.entity_from_eid(vcsrepoeid))
        transaction = vcsrepohdlr.revision_transaction(session, entity)
        transaction.reveid = entity.eid
        transactions[vcsrepoeid] = transaction
        entity.cw_edited['revision'] = transaction.rev
        # VCTransactionOp is a late operation, but we shouldn't wait that much
        # to set at_revision relation, so use two distinct operation
        VCSetAtRevisionOp(session, revision=entity)
        VCTransactionOp(session, revision=entity, transaction=transaction)


class AddVersionContentHook(hook.Hook):
    __regid__ = 'vcsfile.add_version_content_hook'
    __select__ = hook.Hook.__select__ & is_instance('VersionContent')
    events = ('before_add_entity',)
    category = 'metadata'

    def __call__(self):
        edited = self.entity.cw_edited
        if need_guess(edited.get('data_format'), edited.get('data_encoding')):
            vf = self.entity._vc_vf()
            encoding = vf.repository.encoding
            mt, enc = guess_mimetype_and_encoding(data=edited.get('data'),
                                                  filename=vf.name,
                                                  fallbackencoding=encoding)
            if mt and not edited.get('data_format'):
                edited['data_format'] = unicode(mt)
            if enc and not edited.get('data_encoding'):
                edited['data_encoding'] = unicode(enc)


class AddDeletedVersionContentHook(hook.Hook):
    __regid__ = 'vcsfile.add_deleted_version_content_hook'
    __select__ = (hook.Hook.__select__ & is_instance('DeletedVersionContent')
                  & session.is_user_session())
    events = ('before_add_entity',)
    # no category, mandatory hook

    def __call__(self):
        vcsrepohdlr, transaction = self.entity._vc_prepare()
        vf = self.entity._vc_vf()
        vcsrepohdlr.add_versioned_file_deleted_content(self._cw, transaction, vf,
                                                       self.entity)


class AddVersionedFileHook(hook.Hook):
    __regid__ = 'vcsfile.add_versioned_file_hook'
    __select__ = (hook.Hook.__select__ & is_instance('VersionedFile')
                  & session.is_user_session())
    events = ('before_add_entity',)
    category = 'integrity'

    def __call__(self):
        CheckVersionedFileOp(self._cw, entity=self.entity)

### Other hooks
###############################################################################

class SetRepoCacheBeforeAddHook(hook.Hook):
    """clone just after repo creation

    mostly helps the tests
    """
    __regid__ = 'clone_repo_after_add'
    __select__ = (hook.Hook.__select__ & is_instance('Repository'))
    events = ('before_add_entity',)
    category = 'vcsfile.cache'
    order = 2

    def __call__(self):
        repo = self.entity
        if repo.type == 'mercurial' and repo.source_url and not repo.path:
            set_local_cache(repo)


class UpdateRepositoryHook(hook.Hook):
    """add repository eid to vcs bridge cache

    - Check that the type attribute is immutable
    - refresh the cache on url changes

    XXX stuff claiming to be cache but actually necessary are EVIL"""
    __regid__ = 'vcsfile.update_repository_hook'
    __select__ = hook.Hook.__select__ & is_instance('Repository')
    events = ('before_update_entity', )
    # XXX no category, contains integrity, mandatory and vcsfile.cache behaviour

    def __call__(self):
        vcsrepo = self.entity
        if 'type' in vcsrepo.cw_edited:
            msg = self._cw._('updating type attribute of a repository isn\'t '
                             'supported. Delete it and add a new one.')
            raise ValidationError(vcsrepo.eid, {'type': msg})
        if 'path' in vcsrepo.cw_edited or 'local_cache' in vcsrepo.cw_edited:
            bridge._REPOHDLRS.pop(vcsrepo.eid, None)
        if (vcsrepo.type == 'mercurial' and vcsrepo.source_url
            and not vcsrepo.path and not vcsrepo.local_cache):
            # XXX rmdir on rollback
            set_local_cache(vcsrepo)
            if (vcsrepo.import_revision_content
                or self._cw.vreg.config['repository-import']):
                clone_to_local_cache(vcsrepo)


class DeleteDirsOp(hook.DataOperationMixIn, hook.Operation):

    def postcommit_event(self):
        for cachedir in self.get_data():
            try:
                try:
                    shutil.rmtree(cachedir)
                    self.info('deleted repository cache at %s', cachedir)
                except OSError, exc:
                    if (exc.errno != errno.ENOENT
                        or getattr(exc, 'filename', None) != cachedir):
                        raise
            except Exception:
                self.exception('cannot delete repository cache at %s', cachedir)

class DeleteRepositoryHook(hook.Hook):
    __regid__ = 'vcsfile.add_update_repository_hook'
    __select__ = (hook.Hook.__select__ &
                  is_instance('Repository') &
                  score_entity(lambda x:x.local_cache))
    events = ('before_delete_entity',)
    # XXX mandatory hook?

    def __call__(self):
        if self.entity.local_cache:
            cachedir = osp.dirname(osp.join(
                self._cw.vreg.config.local_repo_cache_root,
                self.entity.local_cache)).encode('ascii')
            DeleteDirsOp.get_instance(self._cw).add_data(cachedir)


# safety belts #################################################################

def _check_in_transaction(vf_or_rev):
    """check that a newly added versioned file or revision entity is done in
    a vcs repository transaction.
    """
    try:
        vcsrepo = vf_or_rev.from_repository[0]
    except IndexError:
        raise missing_relation_error(vf_or_rev, 'from_repository')
    try:
        transactions = vcsrepo._cw.transaction_data['vctransactions']
        transaction = transactions[vcsrepo.eid]
    except KeyError:
        raise QueryError('no transaction in progress for repository %s'
                         % vcsrepo.eid)
    return transaction


class CheckVersionedFileOp(hook.Operation):
    """check transaction consistency when adding new revision using rql queries
    """
    def precommit_event(self):
        _check_in_transaction(self.entity)


class CheckRevisionOp(hook.Operation):
    """check transaction consistency when adding new revision using rql queries
    """
    def precommit_event(self):
        try:
            revision = self.entity.from_revision[0]
        except IndexError:
            raise missing_relation_error(self.entity, 'from_revision')
        transaction = _check_in_transaction(revision)
        if not transaction.reveid == revision.eid:
            raise QueryError('entity linked to a bad revision')


class CheckImmutableAttributeHook(hook.Hook):
    __regid__ = 'vcsfile.check_immutable_attribute_hook'
    __select__ = (hook.Hook.__select__
                  & is_instance('Revision', 'DeletedVersionContent', 'VersionContent')
                  & session.is_user_session())
    events = ('before_update_entity',)
    category = 'integrity'

    def __call__(self):
        for attr in self.entity.cw_edited:
            if attr == 'eid':
                continue
            if '%s.%s' % (self.entity.__regid__, attr) in IMMUTABLE_ATTRIBUTES:
                raise QueryError('%s attribute is not editable' % attr)


class AddUpdateRepositoryHook(hook.Hook):
    __regid__ = 'vcsfile.add_update_repository_hook'
    __select__ = hook.Hook.__select__ & is_instance('Repository')
    events = ('after_add_entity', 'after_update_entity')
    category = 'integrity'
    required_attrs = ('path', 'source_url')
    order = 1

    def __call__(self):
        entity = self.entity
        if entity.path and not osp.isabs(self.entity.path):
            msg = self._cw._('path must be absolute')
            raise ValidationError(self.entity.eid, {'path': msg})
        if entity.type == 'subversion' and not entity.path:
            msg = self._cw._('path is mandatory for subversion repositories')
            raise ValidationError(entity.eid, {'path': msg})
        if not entity.source_url and not entity.path:
            raise missing_attribute_error(entity, self.required_attrs)
        if entity.source_url and not entity.path and not entity.local_cache:
            try:
                url_to_relative_path(entity.source_url)
            except Exception:
                msg = self._cw._('url must be of the form protocol://path/to/stuff')
                raise ValidationError(entity.eid, {'source_url': msg})

# folder/tag extensions ########################################################

class ClassifyVersionedFileHook(hook.Hook):
    """classifies VersionedFile automatically according to their path in the
    repository (require cubicweb-tag and/or cubicweb-folder installed)
    """
    __regid__ = 'vcsfile.classify_versioned_file_hook'
    __select__ = hook.Hook.__select__ & is_instance('VersionedFile')
    events = ('after_add_entity', )
    category = 'autoset'

    def __call__(self):
        try:
            rschema = self._cw.vreg.schema['filed_under']
            support_folders = (self.entity.e_schema, 'Folder') in rschema.rdefs
        except KeyError:
            support_folders = False
        if not support_folders:
            return
        for directory in self.entity.directory.split(os.sep):
            if not directory:
                continue
            rset = self._cw.execute('Folder X WHERE X name %(name)s',
                                   {'name': directory})
            if rset:
                self._cw.execute('SET X filed_under F WHERE X eid %(x)s, F eid %(f)s',
                                {'x': self.entity.eid, 'f': rset[0][0]})


def _refresh_repos(repo, eids=None, bridge=bridge):
    '''
    Refresh vcs repos designated by the `eids` parameter, if refreshable:

      * None for all vcs repos
      * a list of integers for specific vcs repos
    '''
    with repo.internal_session() as session:
        config = session.vreg.config
        hgrcpath = repo.config['hgrc-path']
        if hgrcpath is not None:
            os.environ['HGRCPATH'] = hgrcpath
        rql = ('Any R, RT, RE, RLC, RSU, RP, RIRC WHERE R is Repository, '
               'NOT (R local_cache NULL AND R path NULL),'
               'R type RT, R encoding RE, R local_cache RLC, R source_url RSU, R path RP,'
               'R import_revision_content RIRC')
        if not config['cache-external-repositories']:
            rql += ', R cw_source S, S name "system"'
        kwargs = None
        if eids is not None:
            if isinstance(eids, int):
                rql += ', R eid %(x)s'
                kwargs = {'x': eids}
            else:
                rql += ', R eid IN (%s)' % ','.join(str(eid) for eid in eids)
        reposrset = session.execute(rql, kwargs)
        for vcsrepo in reposrset.entities():
            # watch for repository being shut-down, our thread may:
            # * not receive the KeyboardInterrupt or shutdown signal
            # * loop a while without accessing the session (actually
            #   until some repository is actually not up-to-date, while
            #   this access to the session would abort the task)
            try:
                repohdlr = bridge.repository_handler(vcsrepo)
            except bridge.VCSException, ex:
                session.error(str(ex))
                continue
            except Exception:
                session.exception('error while retrieving handler for %s',
                                  vcsrepo.eid)
                continue
            if (config['repository-import'] or vcsrepo.import_revision_content) \
                   and vcsrepo.local_cache is not None:
                try:
                    repohdlr.pull_or_clone_cache(vcsrepo.source_url)
                except bridge.VCSException, ex:
                    repohdlr.error(str(ex))
                except Exception:
                    repohdlr.exception(
                        'error while updating local cache of repository %s',
                        vcsrepo.eid)
            if (config['repository-import']
                and vcsrepo.cw_metainformation()['source']['uri'] == 'system'):
                bridge.import_vcsrepo_content(
                    session, repohdlr, vcsrepo,
                    commitevery=config.get('check-revision-commit-every', 1))


class RefreshRepoService(Service):
    __regid__ = 'refresh_repository'
    __select__ = Service.__select__ & match_user_groups('managers')

    def call(self, eids=None):
        _refresh_repos(self._cw.repo, eids)


class VCSRepoNotifHook(hook.Hook):
    events = ('server_startup',)
    __regid__ = 'vcsfile_repo_notif_startup'
    order = 10 # make sure it is loaded "late" (in fact, it must be
               # loaded after ZMQStartHook but we lack a dependency
               # system here, see http://www.cubicweb.org/ticket/2900987)
    def __call__(self):
        def callback(msg):
            self.debug('received notification: %s', ' '.join(msg))
            msg = msg[-1]
            with self.internal_session() as session:
                for suffix in ('', '/'):
                    rset = sessions.execute('Any R WHERE R source_url U, '
                                            'R source_url LIKE %(u)s',
                                            {'u': '%/'+msg+suffix})
                    if rset:
                        break
                if rset:
                    assert len(rset) == 1, 'found multiple repos for "%s"'%msg
                    eid = rset[0][0]
                    self.debug('Calling refresh service for repo %s', eid)
                    session.call_service('refresh_repository', eids=[eid])
                else:
                    self.debug('No repo for %s', msg)

        self.repo.app_instances_bus.add_subscription('vcsfile-repo-notif', callback)



import os
from shutil import rmtree
from cubicweb import UnknownEid

cacheroot = session.repo.config['local-repo-cache-root']
for eid in os.listdir(cacheroot):
    try:
        repo = session.entity_from_eid(int(eid))
    except ValueError:
        print eid, 'is not a valid eid'
    except UnknownEid:
        print 'removing unknown', eid
        rmtree(os.path.join(cacheroot, eid))
    else:
        if not repo.__regid__ == 'Repository':
            print 'removing not-a-repository', eid
            rmtree(os.path.join(cacheroot, eid))

